/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 The ConQAT Project                                   |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package org.conqat.ide.commons.ui.xmlmodel.databinding;

import java.lang.reflect.Method;

import org.eclipse.core.databinding.observable.Diffs;
import org.eclipse.core.databinding.observable.Realm;
import org.eclipse.core.databinding.observable.value.AbstractObservableValue;
import org.eclipse.core.databinding.observable.value.IObservableValue;

import org.conqat.lib.commons.string.StringUtils;
import org.conqat.ide.commons.ui.CCSMCommonsActivator;
import org.conqat.ide.commons.ui.logging.LoggingUtils;
import org.conqat.ide.commons.ui.xmlmodel.IModelListener;
import org.conqat.ide.commons.ui.xmlmodel.XmlRootModelBase;

/**
 * Implementation of {@link IObservableValue} for CCTS/EMF models.
 * 
 * @author hummelb
 * @author $Author: deissenb $
 * @version $Rev: 34252 $
 * @levd.rating GREEN Hash: EA167C1FBEEB66FB5F1F2D37959704B8
 */
public class XmlModelObservableValue extends AbstractObservableValue implements
		IModelListener {

	/** The wrapped object. */
	private final Object element;

	/** The model. */
	private final XmlRootModelBase model;

	/** The getter method used. */
	private final Method getter;

	/** The getter method used. */
	private final Method setter;

	/** The old value. */
	private Object oldValue;

	/** Flag used for detecting our own updates. */
	private boolean updating = false;

	/** Constructor. */
	public XmlModelObservableValue(Realm realm, Object element,
			XmlRootModelBase model, String propertyName)
			throws SecurityException, NoSuchMethodException {
		super(realm);
		this.element = element;
		this.model = model;

		String suffix = StringUtils.capitalize(propertyName);
		getter = element.getClass().getMethod("get" + suffix);
		setter = element.getClass().getMethod("set" + suffix,
				getter.getReturnType());

		oldValue = doGetValue();
	}

	/** {@inheritDoc} */
	@Override
	protected void firstListenerAdded() {
		model.addModelListener(this);
		super.firstListenerAdded();
	}

	/** {@inheritDoc} */
	@Override
	protected void lastListenerRemoved() {
		model.removeModelListener(this);
		super.lastListenerRemoved();
	}

	/** {@inheritDoc} */
	@Override
	public synchronized void dispose() {
		model.removeModelListener(this);
		super.dispose();
	}

	/** {@inheritDoc} */
	@Override
	protected Object doGetValue() {
		try {
			return getter.invoke(element);
		} catch (Exception e) {
			logInvalidMethodCallError(e);
			return null;
		}
	}

	/** {@inheritDoc} */
	@Override
	protected void doSetValue(final Object value) {
		updating = true;
		try {
			setter.invoke(element, value);
		} catch (Exception e) {
			logInvalidMethodCallError(e);
		} finally {
			updating = false;
		}
	}

	/** Log error */
	private void logInvalidMethodCallError(Exception e) {
		LoggingUtils.error(CCSMCommonsActivator.getDefault(),
				"Invalid method call in binding", e);
	}

	/** {@inheritDoc} */
	public Object getValueType() {
		return getter.getReturnType();
	}

	/** {@inheritDoc} */
	public void modelChanged() {
		if (!updating) {
			Object newValue = doGetValue();
			fireValueChange(Diffs.createValueDiff(oldValue, newValue));
			oldValue = newValue;
		}
	}
}